;(load "~/scheme/lib/trace.scm")

(define *dialect* #f)
(define *operating-system* #f)
(define *compile?* #f)
(define *shell-script?* #f)

(define *cr* #f)
(define *lf* #f)

(define string->lower-case
  (lambda (s)
    (list->string
     (map char-downcase
          (string->list s)))))

(define read-a-line
  (lambda (i)
    (list->string
     (let loop ()
       (let ((c (read-char i)))
         (if (or (eof-object? c)
                 (char=? c #\newline))
             '()
             (cons c (loop))))))))

(define copy-file-to-port
  (lambda (f o)
    (call-with-input-file f
      (lambda (i)
        (copy-port-to-port i o)))))

(define copy-port-to-port
  (lambda (i o)
    (let loop ()
      (let ((c (read-char i)))
        (if (not (eof-object? c))
          (begin
            (if (eqv? *operating-system* 'windows)
                (cond ((char=? c *cr*) 'skip)
                      ((char=? c *lf*) (display *cr* o)
                                       (display *lf* o))
                      (else (display c o)))
                (display c o))
            (loop)))))))

(define define-macro?
  (lambda (x)
    (and (pair? x) (eqv? (car x) 'define-macro))))

(define name-defined
  (lambda (x)
    (if (pair? x)
        (let ((n (length x)))
          (cond ((and (> n 2) (memv (car x) '(define define-macro defmacro)))
                 (let ((name (cadr x)))
                   (if (pair? name) (car name) name)))
                ((and (> n 3) (eqv? (car x) 'syntax-table-define))
                 (cadr (caddr x)))
                (else #f)))
        #f)))

(define *files-to-be-ported*
  (call-with-input-file "files-to-be-ported.scm"
    (lambda (i)
      (let loop ()
        (let ((f (read i)))
          (if (eof-object? f) '()
              (cons (string->lower-case (symbol->string f))
                    (loop))))))))

(define *dialects-supported*
  (call-with-input-file "dialects-supported.scm"
    (lambda (i)
      (let loop ()
        (let ((d (read i)))
          (if (eof-object? d) '()
              (cons d (loop))))))))

(define *operating-systems-supported*
  '(unix windows))

(set! *dialect*
  (begin
    (display "What is your Scheme dialect? [")
    (let ((first? #t))
      (for-each
       (lambda (d)
         (if first? (set! first? #f)
             (display " "))
         (display d))
       *dialects-supported*))
    (display "]")
    (newline)
    (read)))

(set! *operating-system*
  (case *dialect*
    ((gambit) 
     (if (getenv "COMSPEC" #f)
	 'windows
	 'unix))
    ((bigloo guile mzscheme petite scm stk)
     (cond ((getenv "COMSPEC") 'windows)
           (else 'unix)))
    ((mitscheme)
     (cond ((get-environment-variable "COMSPEC") 'windows)
           (else 'unix)))
    ((pscheme) 'windows)
    (else
      (display "What is your operating system? [unix windows]")
      (newline)
      (read))))

(set! *compile?*
  (case *operating-system*
    ((unix) (case *dialect*
              ((mzscheme)
               (display "Compile? [#t #f]")
               (newline)
               (read))
              (else #f)))
    (else #f)))

(if (eqv? *operating-system* 'windows)
    (begin
      (set! *cr* (integer->char 13))
      (set! *lf* (integer->char 10))))

(if (and (eqv? *dialect* 'mzscheme)
         *compile?*)
    (require-library "compile.ss"))

(define *dialect-s*
  (string->lower-case (symbol->string *dialect*)))

(define *operating-system-s*
  (string->lower-case (symbol->string *operating-system*)))

(define exists-file?
  (case *dialect*
    ((bigloo gambit guile kawa mitscheme mzscheme petite scm stk)
     file-exists?)
    ((pscheme)
     (lambda (f)
       (with-handlers (((lambda (x) #t) (lambda (x) #f)))
         (close-input-port (open-input-file f))
         #t)))
    (else (lambda (f) #t))))

(define obliterate-file
  (case *dialect*
    ((bigloo guile kawa mitscheme mzscheme petite pscheme scm) delete-file)
    ((gambit)
     ;## causes problems with other Scheme dialects
     (lambda (f)
       ((eval (call-with-input-string "##shell-command" read))
        (string-append "rm " f))))
    (else (lambda (f) #t))))

(define ensure-file-deleted
  (lambda (f)
    (if (exists-file? f) (obliterate-file f))))

(define translate-define-macro
  (case *dialect*
    ((stk)
     (lambda (e)
       `(define-macro (,(cadr e) . ,(cadr (caddr e)))
          ,@(cddr (caddr e)))))
    ((mitscheme)
     (lambda (e)
       `(syntax-table-define system-global-syntax-table ',(cadr e)
          (macro ,@(cdr (caddr e))))))
    ((scm kawa)
     (lambda (e)
       `(defmacro ,(cadr e) ,(cadr (caddr e))
          ,@(cddr (caddr e)))))
    ((gambit)
     (lambda (e)
       `(begin ,e
          (eval ',e))))
    (else ;guile, bigloo, pscheme, etc.
      (lambda (e) e))))

(if (eqv? *dialect* 'petite)
    ;#' causes problems with other dialects
    (begin
      (if (file-exists? ".petite-macro.scm") (delete-file ".petite-macro.scm"))
      (call-with-output-file ".petite-macro.scm"
        (lambda (o)
          (display "
(set! translate-define-macro
  (lambda (e)
    `(define-syntax ,(cadr e)
       (lambda (x)
         (syntax-case x ()
           ((k . stuff)
            (datum->syntax-object #'k
              (apply ,(caddr e)
                     (cdr (syntax-object->datum x))))))))))
" o)))))

(if (eqv? *dialect* 'petite)
    (load ".petite-macro.scm"))

(cond ((not (memv *dialect* *dialects-supported*))
       (display "Sorry, dialect ")
       (display *dialect*)
       (display " is not supported. :-<")
       (newline))
      ((not (memv *operating-system* *operating-systems-supported*))
       (display "Sorry, operating system ")
       (display *operating-system*)
       (display " is not supported.")
       (newline))
      (else
       (for-each
        (lambda (file-to-be-ported)
          (set! *shell-script?* #f)
          (display "Porting ")
          (display file-to-be-ported)
          (display "...")
          (newline)
          (let ((preamble1 (string-append *dialect-s* "-preamble-"
                                          file-to-be-ported))
                (preamble2 (string-append *dialect-s* "-"
                                          *operating-system-s*
                                          "-preamble-"
                                          file-to-be-ported))
                (procs1 (string-append *dialect-s* "-procs-"
                                       file-to-be-ported))
                (procs2 (string-append *dialect-s* "-"
                                       *operating-system-s*
                                       "-procs-"
                                       file-to-be-ported))
                (postamble1 (string-append *dialect-s* "-postamble-"
                                           file-to-be-ported))
                (postamble2 (string-append *dialect-s* "-"
                                           *operating-system-s*
                                           "-postamble-"
                                           file-to-be-ported)))
            (let ((preamble-file
                   (cond ((exists-file? preamble2) preamble2)
                         ((exists-file? preamble1) preamble1)
                         (else #f)))
                  (procs-file-generic
                   (if (exists-file? procs1) procs1 #f))
                  (procs-file-system-specific
                   (if (exists-file? procs2) procs2 #f))
                  (postamble-file
                   (cond ((exists-file? postamble2) postamble2)
                         ((exists-file? postamble1) postamble1)
                         (else #f)))
                  (target-file
                   (string-append *dialect-s* "-"
                                  *operating-system-s*
                                  "-version-of-"
                                  file-to-be-ported)))
              (ensure-file-deleted target-file)
              (call-with-output-file target-file
                (lambda (o)
                  (if (and preamble-file (not *compile?*))
                      (begin
                        (copy-file-to-port preamble-file o)
                        (newline o)))
                  (let ((names-defined '()))
                    (if procs-file-generic
                        (call-with-input-file procs-file-generic
                          (lambda (i)
                            (let loop ()
                              (let ((x (read i)))
                                (if (not (eof-object? x))
                                    (let ((name (name-defined x)))
                                      (if name
                                          (set! names-defined
                                                (cons name names-defined)))
                                      (pp x o)
                                      (newline o)
                                      (loop))))))))
                    (if procs-file-system-specific
                        (call-with-input-file procs-file-system-specific
                          (lambda (i)
                            (let loop ()
                              (let ((x (read i)))
                                (if (not (eof-object? x))
                                    (let ((name (name-defined x)))
                                      (if name
                                          (set! names-defined
                                                (cons name names-defined)))
                                      (pp x o)
                                      (newline o)
                                      (loop))))))))
                    (call-with-input-file file-to-be-ported
                      (lambda (i)
                        (if (char=? (peek-char i) #\#)
                            (begin
                              (set! *shell-script?* #t)
                              (read-a-line i)))
                        (let loop ()
                          (let ((x (read i)))
                            (if (not (eof-object? x))
                                (begin
                                  (if (not (memv (name-defined x)
                                                 names-defined))
                                      (begin
                                        (pp
                                         (if (define-macro? x)
                                             (translate-define-macro x)
                                             x) o)
                                        (newline o)))
                                  (loop))))))))
                  (newline o)
                  (if postamble-file
                      (copy-file-to-port postamble-file o))
                  (newline o)))
              (if *compile?*
                  (let ((target-file-so
                         (string-append target-file ".so")))
                    (display "Compiling...")
                    (newline)
                    (case *dialect*
                      ((mzscheme)
                       (compile-file target-file target-file-so)
                       (delete-file target-file)
                       (call-with-output-file target-file
                         (lambda (o)
                           (if preamble-file
                               (begin
                                 (copy-file-to-port preamble-file o)
                                 (newline o)))
                           (copy-file-to-port target-file-so o)))))))
              (if *shell-script?*
                  (if (and (memv *dialect*
                                 '(guile mitscheme petite))
                           (eqv? *operating-system* 'unix))
                      (system
                       (string-append "chmod +x "
                                      target-file))))
              (display "Resulting file is `")
              (display target-file)
              (display "'")
              (newline)
              (display "You may want to rename it")
              (newline)
              )))
        *files-to-be-ported*)))

(case *dialect*
  ((bigloo) (exit 0))
  ((gambit guile mzscheme petite pscheme stk) (exit))
  ((mitscheme) (%exit))
  ((scm) (quit))
  (else (display "You may exit Scheme now!")
        (newline)))
